/*
 * Copyright (C) 2004-2025 ZNC, see the NOTICE file for details.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include <znc/ZNCDebug.h>

TEST(DebugTest, Filter) {
    EXPECT_EQ(CDebug::Filter("JOIN #znc"), "JOIN #znc");
    EXPECT_EQ(CDebug::Filter("PASS hunter2"), "PASS <censored>");
    EXPECT_EQ(CDebug::Filter("PASS :hunter2"), "PASS :<censored>");
    EXPECT_EQ(CDebug::Filter("PASS user/net:hunter2"), "PASS user/net:<censored>");
    EXPECT_EQ(CDebug::Filter("PASS :user/net:hunter2"), "PASS :user/net:<censored>");
    EXPECT_EQ(CDebug::Filter("pass hunter2"), "pass <censored>");
}
